import datetime

import pytest

from d8s_domains import (
    domain_as_punycode,
    domain_as_unicode,
    domain_certificate_peers,
    domain_dns,
    domain_examples,
    domain_rank,
    domain_second_level_name,
    domain_subdomains,
    domain_tld,
    domain_whois,
    domains_find,
    is_domain,
    is_tld,
    tlds,
)


def test_domain_certificate_peers_docs_1():
    assert 'android.com' in domain_certificate_peers('google.com')
    # make sure that the domain name is parsed from a URL when a URL is given to the domain_certificate_peers function
    assert 'android.com' in domain_certificate_peers('https://google.com/')


def test_domain_examples_docs_1():
    assert len(domain_examples()) == 10
    assert len(domain_examples(n=101)) == 101


def test_domains_find_docs_1():
    s = '''foo.com
    bingo.co.uk
    BAR.COM'''
    domain_list = domains_find(s)
    assert len(domain_list) == 3
    assert 'foo.com' in domain_list
    assert 'bingo.co.uk' in domain_list
    assert 'bar.com' in domain_list


@pytest.mark.network
def test_domain_rank_docs_1():
    domain = 'example.com'
    assert domain_rank(domain) < 10000


@pytest.mark.network
def test_domain_rank_given_url():
    domain = 'https://example.com/test/bingo.php'
    assert domain_rank(domain) < 10000


@pytest.mark.network
def test_tlds_docs_1():
    assert len(tlds()) > 1500


def test_domain_as_punycode_docs_1():
    assert domain_as_punycode('☁.com') == 'xn--l3h.com'
    assert domain_as_punycode('ドメイン.テスト') == 'xn--eckwd4c7c.xn--zckzah'
    assert domain_as_punycode('http://例子.卷筒纸') == 'xn--fsqu00a.xn--3lr804guic'
    assert domain_as_punycode('example.com') == 'example.com'
    assert domain_as_punycode('Ἐν ἀρχῇ ἦν ὁ λόγος') == 'xn--    -emd6arkdyo6a6bye7519ciharniqk5f'


def test_domain_as_unicode_docs_1():
    assert domain_as_unicode('xn--l3h.com') == '☁.com'
    assert domain_as_unicode('https://xn--l3h.com/test/bingo.php') == '☁.com'


def test_domain_dns_docs_1():
    assert domain_dns('example.com') == '93.184.216.34'


def test_domain_second_level_name_docs_1():
    assert domain_second_level_name('http://example.com/test/bingo') == 'example'
    assert domain_second_level_name('example.com') == 'example'
    assert domain_second_level_name('forums.bbc.co.uk') == 'bbc'
    assert domain_second_level_name('http://www.worldbank.org.kg/') == 'worldbank'


def test_domain_subdomains_docs_1():
    assert domain_subdomains('https://foo.bar.example.com') == 'foo.bar'
    assert domain_subdomains('https://foo.bar.example.co.uk') == 'foo.bar'
    assert domain_subdomains('foo.bar.google.co.uk') == 'foo.bar'
    assert domain_subdomains('http://www.worldbank.org.kg/') == 'www'


def test_domain_tld_docs_1():
    assert domain_tld('google.com') == 'com'
    assert domain_tld('forums.bbc.co.uk') == 'co.uk'
    assert domain_tld('http://www.worldbank.org.kg/') == 'org.kg'


def test_domain_whois_docs_1():
    assert domain_whois('google.com') == {
        'name': 'google.com',
        'registrar': 'MarkMonitor Inc.',
        'creation_date': datetime.datetime(1997, 9, 15, 4, 0),
        'expiration_date': datetime.datetime(2028, 9, 14, 4, 0),
        'last_updated': None,
        'status': 'clientDeleteProhibited https://icann.org/epp#clientDeleteProhibited',
        'name_servers': {
            'ns2.google.com',
            'ns2.google.com\r',
            'ns1.google.com\r',
            'ns4.google.com',
            'ns3.google.com\r',
            'ns4.google.com\r',
            'ns3.google.com',
            'ns1.google.com',
        },
    }
    assert domain_whois('example.com') is None


def test_is_domain_docs_1():
    assert is_domain('example.com')
    assert is_domain('forums.bbc.co.uk')
    assert is_domain('foo.bar')
    assert not is_domain('foo')


def test_is_tld_docs_1():
    assert is_tld('com')
    assert is_tld('org')
    assert not is_tld('foobar')
    assert not is_tld('ocx')
    assert not is_tld('dotm')
    assert is_tld('.com')



import socket
import ssl
from typing import Any, Dict, List, Optional

import tldextract
from d8s_hypothesis import hypothesis_get_strategy_results
from d8s_networking import get
from hypothesis.provisional import domains
from ioc_finder import ioc_finder

from .domains_temp_utils import get_first_arg_url_domain


def is_domain(possible_domain: str) -> bool:
    """Check if the given string is a domain."""
    domains = domains_find(possible_domain)
    if len(domains) == 1 and domains[0] == possible_domain:
        return True

    return False


# TODO: this function will sometimes create/return duplicates... probably need to have some option to prevent duplicates
def domain_examples(n: int = 10) -> List[str]:
    """Create n domain names."""
    results = hypothesis_get_strategy_results(domains, n=n)

    return results


def domains_find(text: str, **kwargs: bool) -> List[str]:
    """Parse domain names from the given text."""
    return ioc_finder.parse_domain_names(text, **kwargs)


@get_first_arg_url_domain
def domain_dns(domain: str) -> str:
    """Get the DNS results for the given domain."""
    return socket.gethostbyname(domain)


@get_first_arg_url_domain
def domain_certificate_peers(domain: str) -> List[str]:
    """Return a list of all domains sharing a certificate with the given domain."""
    ctx = ssl.create_default_context()
    s = ctx.wrap_socket(socket.socket(), server_hostname=domain)
    s.connect((domain, 443))
    cert = s.getpeercert()
    return [domain_name[1] for domain_name in cert['subjectAltName']]  # type: ignore


@get_first_arg_url_domain
def domain_whois(domain: str) -> Optional[Dict[str, Any]]:
    """."""
    import whois

    whois_data = whois.query(domain)
    if whois_data:
        return whois_data.__dict__
    else:
        return None


@get_first_arg_url_domain
def domain_subdomains(domain_name: str) -> str:
    """Get the subdomains for the given domain name."""
    return tldextract.extract(domain_name).subdomain


@get_first_arg_url_domain
def domain_second_level_name(domain_name: str) -> str:
    """Get the second level name for the given domain name (e.g. google from https://google.co.uk)."""
    return tldextract.extract(domain_name).domain


@get_first_arg_url_domain
def domain_tld(domain_name: str) -> str:
    """Get the top level domain for the given domain name."""
    return tldextract.extract(domain_name).suffix


@get_first_arg_url_domain
def domain_rank(domain_name: str) -> int:
    """."""
    onemillion_api_url = f'http://onemillion.hightower.space/onemillion/{domain_name}'
    return get(onemillion_api_url, process_response=True)


@get_first_arg_url_domain
def domain_as_punycode(domain_name: str) -> str:
    """Convert the given domain name to Punycode (https://en.wikipedia.org/wiki/Punycode)."""
    return domain_name.encode('idna').decode('utf-8')


@get_first_arg_url_domain
def domain_as_unicode(domain_name: str) -> str:
    """Convert a given domain name to Unicode (https://en.wikipedia.org/wiki/Unicode)."""
    return domain_name.encode('utf-8').decode('idna')


# TODO: cache this data
def tlds() -> List[str]:
    """Get the top level domains from https://iana.org/."""
    top_level_domains = get('https://data.iana.org/TLD/tlds-alpha-by-domain.txt', process_response=True).split('\n')[
        1:-1
    ]
    return [tld.lower() for tld in top_level_domains]


def is_tld(possible_tld: str) -> bool:
    """Return whether or not the possible_tld is a valid tld."""
    # remove any periods from the beginning (in case someone provides the TLD like `.com` rather than just `com`)
    possible_tld = possible_tld.lstrip('.')

    valid_tlds = [tld.lower() for tld in tlds()]
    tld_is_valid = possible_tld.lower() in valid_tlds
    return tld_is_valid



import functools


def get_first_arg_url_domain(func):
    """If the first argument is a url, get the domain of the url and pass that into the function."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        from d8s_urls import is_url, url_domain

        domain_arg = args[0]
        other_args = args[1:]

        if is_url(domain_arg):
            domain_arg = url_domain(domain_arg)

        return func(domain_arg, *other_args, **kwargs)

    return wrapper
